@@ -2,7 +2,7 @@
  *
  *  Wireless daemon for Linux
  *
- *  Copyright (C) 2017-2019  Intel Corporation. All rights reserved.
+ *  Copyright (C) 2017-2020  Intel Corporation. All rights reserved.
  *
  *  This library is free software; you can redistribute it and/or
  *  modify it under the terms of the GNU Lesser General Public
@@ -49,7 +49,8 @@ static struct l_queue *proxy_interface_types;
 
 void proxy_properties_display(const struct proxy_interface *proxy,
 				const char *caption, const char *margin,
-				int name_column_width, int value_column_width)
+				unsigned int name_column_width,
+				unsigned int value_column_width)
 {
 	const void *data;
 	const struct proxy_interface_property *properties;
@@ -58,7 +59,7 @@ void proxy_properties_display(const struct proxy_interface *proxy,
 	if (!proxy->type->properties)
 		return;
 
-	display_table_header(caption, "%s%-*s  %-*s%-*s", margin,
+	display_table_header(caption, "%s%-*s  %-*s  %-*s", margin,
 				8, "Settable",
 				name_column_width, "Property",
 				value_column_width, "Value");
@@ -67,14 +68,17 @@ void proxy_properties_display(const struct proxy_interface *proxy,
 	properties = proxy->type->properties;
 
 	for (i = 0; properties[i].name; i++) {
+		const char *str;
+
 		if (!properties[i].tostr)
 			continue;
 
-		display("%s%*s  %-*s%-.*s\n", margin,
-			8, properties[i].is_read_write ?
-				COLOR_BOLDGRAY "       *" COLOR_OFF : "",
-			name_column_width, properties[i].name,
-			value_column_width, properties[i].tostr(data) ? : "");
+		str = properties[i].tostr(data);
+
+		display_table_row(MARGIN, 3, 8, properties[i].is_read_write ?
+				COLOR_BOLDGRAY("       *") : "",
+				name_column_width, properties[i].name,
+				value_column_width, str ? : "");
 	}
 }
 
@@ -107,6 +111,9 @@ static void proxy_interface_property_update(struct proxy_interface *proxy,
 	const struct proxy_interface_property *property_table =
 							proxy->type->properties;
 
+	if (!property_table)
+		return;
+
 	for (i = 0; property_table[i].name; i++) {
 		if (strcmp(property_table[i].name, name))
 			continue;
@@ -315,7 +322,8 @@ static void proxy_callback(struct l_dbus_message *message, void *user_data)
 	}
 
 	proxy = callback_data->user_data;
-	if (!strcmp(proxy->type->interface, IWD_AGENT_MANAGER_INTERFACE))
+	if (!strcmp(proxy->type->interface, IWD_AGENT_MANAGER_INTERFACE) ||
+			!strcmp(proxy->type->interface, IWD_DAEMON_INTERFACE))
 		return;
 
 quit:
@@ -511,6 +519,33 @@ static bool is_ignorable(const char *interface)
 	return false;
 }
 
+static void proxy_interfaces_update_properties(const char *path,
+					struct l_dbus_message_iter *interfaces)
+{
+	const char *interface;
+	struct l_dbus_message_iter properties;
+	struct proxy_interface *proxy;
+	struct proxy_interface_type *interface_type;
+
+	if (!path)
+		return;
+
+	while (l_dbus_message_iter_next_entry(interfaces, &interface,
+								&properties)) {
+		interface_type = l_queue_find(proxy_interface_types,
+						interface_match_by_type_name,
+						interface);
+		if (!interface_type)
+			continue;
+
+		proxy = proxy_interface_find(interface_type->interface, path);
+		if (!proxy)
+			continue;
+
+		interface_update_properties(proxy, &properties, NULL);
+	}
+}
+
 static void proxy_interface_create(const char *path,
 					struct l_dbus_message_iter *interfaces)
 {
@@ -538,23 +573,17 @@ static void proxy_interface_create(const char *path,
 
 		proxy = proxy_interface_find(interface_type->interface, path);
 
-		if (proxy) {
-			interface_update_properties(proxy, &properties, NULL);
-
+		if (proxy)
 			continue;
-		}
 
 		proxy = l_new(struct proxy_interface, 1);
 		proxy->path = l_strdup(path);
 		proxy->type = interface_type;
 
-		if (interface_type->ops && interface_type->ops->create) {
-			proxy->data = interface_type->ops->create();
-
-			interface_update_properties(proxy, &properties, NULL);
-		}
-
 		l_queue_push_tail(proxy_interfaces, proxy);
+
+		if (interface_type->ops && interface_type->ops->create)
+			proxy->data = interface_type->ops->create();
 	}
 }
 
@@ -656,6 +685,12 @@ static void interfaces_added_callback(struct l_dbus_message *message,
 		return;
 
 	proxy_interface_create(path, &object);
+
+	if (!l_dbus_message_get_arguments(message, "oa{sa{sv}}", &path,
+								&object))
+		return;
+
+	proxy_interfaces_update_properties(path, &object);
 }
 
 static void interfaces_removed_callback(struct l_dbus_message *message,
@@ -692,52 +727,44 @@ static void get_managed_objects_callback(struct l_dbus_message *message,
 	const char *path;
 
 	if (dbus_message_has_error(message)) {
-		l_error("Failed to retrieve IWD dbus objects, quitting...\n");
+		display_error("Failed to retrieve IWD dbus objects, "
+							"quitting...\n");
 
-		if (!command_is_interactive_mode())
-			command_set_exit_status(EXIT_FAILURE);
-
-		l_main_quit();
-
-		return;
+		goto error;
 	}
 
 	if (!l_dbus_message_get_arguments(message, "a{oa{sa{sv}}}", &objects)) {
-		l_error("Failed to parse IWD dbus objects, quitting...\n");
+		display_error("Failed to parse IWD dbus objects, "
+							"quitting...\n");
 
-		if (!command_is_interactive_mode())
-			command_set_exit_status(EXIT_FAILURE);
-
-		l_main_quit();
-
-		return;
+		goto error;
 	}
 
 	while (l_dbus_message_iter_next_entry(&objects, &path, &object))
 		proxy_interface_create(path, &object);
 
-	if (command_needs_no_agent())
-		goto no_agent;
+	if (!l_dbus_message_get_arguments(message, "a{oa{sa{sv}}}", &objects))
+		/*
+		 * Shouldn't happen since we parsed it above, but check return
+		 * anyway.
+		 */
+		goto error;
 
-	if (!agent_manager_register_agent()) {
-		display_error("Failed to register Agent.\n");
+	while (l_dbus_message_iter_next_entry(&objects, &path, &object))
+		proxy_interfaces_update_properties(path, &object);
 
-		if (!command_is_interactive_mode())
-			command_set_exit_status(EXIT_FAILURE);
-
-		l_main_quit();
-
-		return;
-	}
-
-no_agent:
-	if (!command_is_interactive_mode()) {
+	if (command_is_interactive_mode())
+		display_enable_cmd_prompt();
+	else
 		command_noninteractive_trigger();
 
-		return;
-	}
+	return;
 
-	display_enable_cmd_prompt();
+error:
+	if (!command_is_interactive_mode())
+		command_set_exit_status(EXIT_FAILURE);
+
+	l_main_quit();
 }
 
 static void get_managed_objects(void)
@@ -851,8 +878,7 @@ bool dbus_proxy_exit(void)
 {
 	struct interface_type_desc *desc;
 
-	if (!command_needs_no_agent())
-		agent_manager_unregister_agent();
+	agent_manager_unregister_agent();
 
 	for (desc = __start___interface; desc < __stop___interface; desc++) {
 		if (!desc->exit)
