@@ -89,7 +89,8 @@ bool proxy_interface_method_call(const struct proxy_interface *proxy,
 
 void proxy_properties_display(const struct proxy_interface *proxy,
 				const char *caption, const char *margin,
-				int name_column_width, int value_column_width);
+				unsigned int name_column_width,
+				unsigned int value_column_width);
 
 char *proxy_property_str_completion(const struct proxy_interface_type *type,
 					proxy_property_match_func_t function,
