@@ -27,19 +27,23 @@
 #define _GNU_SOURCE
 #include <stdio.h>
 #include <signal.h>
+#include <sys/stat.h>
+#include <sys/ioctl.h>
+#include <unistd.h>
 
 #include <readline/history.h>
 #include <readline/readline.h>
 #include <ell/ell.h>
 
-#include "agent.h"
-#include "command.h"
-#include "display.h"
+#include "client/agent.h"
+#include "client/command.h"
+#include "client/display.h"
 
-#define IWD_PROMPT COLOR_GREEN "[iwd]" COLOR_OFF "# "
+#define IWD_PROMPT \
+	"\001" COLOR_GREEN("\002" "[iwd]" "\001") "\002" "# "
 #define LINE_LEN 81
 
-static struct l_signal *resize_signal;
+static struct l_signal *window_change_signal;
 static struct l_io *io;
 static char dashed_line[LINE_LEN] = { [0 ... LINE_LEN - 2] = '-' };
 static char empty_line[LINE_LEN] = { [0 ... LINE_LEN - 2] = ' ' };
@@ -47,6 +51,7 @@ static struct l_timeout *refresh_timeout;
 static struct saved_input *agent_saved_input;
 
 static struct display_refresh {
+	bool enabled;
 	char *family;
 	char *entity;
 	const struct command *cmd;
@@ -55,7 +60,7 @@ static struct display_refresh {
 	size_t undo_lines;
 	struct l_queue *redo_entries;
 	bool recording;
-} display_refresh;
+} display_refresh = { .enabled = true };
 
 struct saved_input {
 	char *line;
@@ -220,12 +225,45 @@ void display_refresh_set_cmd(const char *family, const char *entity,
 	}
 }
 
+static void display_refresh_check_feasibility(void)
+{
+	const struct winsize ws;
+
+	ioctl(STDOUT_FILENO, TIOCGWINSZ, &ws);
+
+	if (ws.ws_col < LINE_LEN - 1) {
+		if (display_refresh.enabled) {
+			display_refresh.recording = false;
+			display(COLOR_YELLOW("Auto-refresh is disabled. "
+				"Enlarge window width to at least %u to enable."
+				"\n"), LINE_LEN - 1);
+			display_refresh.recording = true;
+		}
+
+		display_refresh.enabled = false;
+	} else {
+		display_refresh.enabled = true;
+	}
+}
+
+static void display_refresh_check_applicability(void)
+{
+	if (display_refresh.enabled && display_refresh.cmd)
+		display_refresh_redo_lines();
+	else if (display_refresh.cmd)
+		display_refresh_timeout_set();
+}
+
 static void timeout_callback(struct l_timeout *timeout, void *user_data)
 {
 	struct saved_input *input;
 
-	if (!display_refresh.cmd)
+	if (!display_refresh.enabled || !display_refresh.cmd) {
+		if (display_refresh.cmd)
+			display_refresh_timeout_set();
+
 		return;
+	}
 
 	input = save_input();
 	display_refresh_undo_lines();
@@ -279,7 +317,7 @@ void display(const char *fmt, ...)
 
 void display_error(const char *error)
 {
-	char *text = l_strdup_printf(COLOR_RED "%s" COLOR_OFF "\n", error);
+	char *text = l_strdup_printf(COLOR_RED("%s\n"), error);
 
 	display_text(text);
 
@@ -306,14 +344,14 @@ void display_table_header(const char *caption, const char *fmt, ...)
 	int caption_pos =
 		(int) ((sizeof(dashed_line) - 1) / 2 + strlen(caption) / 2);
 
-	text = l_strdup_printf("%*s" COLOR_BOLDGRAY "%*c" COLOR_OFF "\n",
+	text = l_strdup_printf("%*s" COLOR_BOLDGRAY("%*c") "\n",
 				caption_pos, caption,
 				LINE_LEN - 2 - caption_pos,
 				display_refresh.cmd ? get_flasher() : ' ');
 	display_text(text);
 	l_free(text);
 
-	text = l_strdup_printf("%s%s%s\n", COLOR_GRAY, dashed_line, COLOR_OFF);
+	text = l_strdup_printf(COLOR_GRAY("%s\n"), dashed_line);
 	display_text(text);
 	l_free(text);
 
@@ -321,12 +359,12 @@ void display_table_header(const char *caption, const char *fmt, ...)
 	text = l_strdup_vprintf(fmt, args);
 	va_end(args);
 
-	body = l_strdup_printf("%s%s%s\n", COLOR_BOLDGRAY, text, COLOR_OFF);
+	body = l_strdup_printf(COLOR_BOLDGRAY("%s\n"), text);
 	display_text(body);
 	l_free(body);
 	l_free(text);
 
-	text = l_strdup_printf("%s%s%s\n", COLOR_GRAY, dashed_line, COLOR_OFF);
+	text = l_strdup_printf(COLOR_GRAY("%s\n"), dashed_line);
 	display_text(text);
 	l_free(text);
 }
@@ -335,12 +373,186 @@ void display_table_footer(void)
 {
 	display_text("\n");
 
-	if (display_refresh.cmd)
-		display_refresh_redo_lines();
+	display_refresh_check_applicability();
 }
 
-void display_command_line(const char *command_family,
-						const struct command *cmd)
+static unsigned int color_end(char *s)
+{
+	char *start = s;
+
+	while (*s != 'm' && *s != '\0')
+		s++;
+
+	return s - start + 1;
+}
+
+/*
+ * Finds last space in 's' before 'max' characters, terminates at that index,
+ * and returns a new string to be printed on the next line.
+ *
+ * 'max' should be set to the column width, but is also an out parameter since
+ * this width can be updated if colored escapes are detected.
+ *
+ * Any colored escapes found are set to 'color_out' so they can be re-enabled
+ * on the next line.
+ */
+static char* next_line(char *s, unsigned int *max, char **color_out)
+{
+	unsigned int i;
+	int last_space = -1;
+	int last_color = -1;
+
+	/* Find the last space before 'max', as well as any color */
+	for (i = 0; i <= *max && s[i] != '\0'; i++) {
+		if (s[i] == ' ')
+			last_space = i;
+		else if (s[i] == 0x1b) {
+			/* color escape won't count for column width */
+			*max += color_end(s + i);
+			last_color = i;
+		}
+	}
+
+	/* Reached the end of the string within the column bounds */
+	if (i <= *max)
+		return NULL;
+
+	/* Not anywhere nice to split the line */
+	if (last_space == -1)
+		last_space = *max - 1;
+
+	/*
+	 * Only set the color if it occurred prior to the last space. If after,
+	 * it will get picked up on the next line.
+	 */
+	if (last_color != -1 && last_space >= last_color)
+		*color_out = l_strndup(s + last_color,
+					color_end(s + last_color));
+	else
+		*color_out = NULL;
+
+	s[last_space] = '\0';
+
+	return l_strdup(s + last_space + 1);
+}
+
+struct table_entry {
+	unsigned int width;
+	char *next;
+	char *color;
+};
+
+/*
+ * Appends the next line from 'e' to 'line_buf'. 'done' is only set false when
+ * there are more lines needed for the current entry.
+ */
+static int entry_append(struct table_entry *e, char *line_buf)
+{
+	char *value = e->next;
+	unsigned int ret = 0;
+	unsigned int width = e->width;
+
+	/* Empty line */
+	if (!value)
+		return sprintf(line_buf, "%-*s  ", e->width, "");
+
+	/* Color from previous line */
+	if (e->color) {
+		ret = sprintf(line_buf, "%s", e->color);
+		l_free(e->color);
+		e->color = NULL;
+	}
+
+	/* Advance entry to next line, and terminate current */
+	e->next = next_line(value, &width, &e->color);
+
+	/* Append current line */
+	ret += sprintf(line_buf + ret, "%-*s  ", width, value);
+
+	l_free(value);
+
+	/* Un-color output for next column */
+	if (e->color)
+		ret += sprintf(line_buf + ret, "%s", COLOR_OFF);
+
+	return ret;
+}
+
+static bool entries_done(unsigned int num, struct table_entry *e)
+{
+	unsigned int i;
+
+	for (i = 0; i < num; i++)
+		if (e[i].next)
+			return false;
+
+	return true;
+}
+
+/*
+ * Expects an initial margin, number of columns in table, then row data:
+ *
+ * <row width>, <row data>, ...
+ *
+ * The data string can be of any length, and will be split into new lines of
+ * length <row width>.
+ */
+void display_table_row(const char *margin, unsigned int ncolumns, ...)
+{
+	char buf[512];
+	char *str = buf;
+	unsigned int i;
+	struct table_entry entries[ncolumns];
+	va_list va;
+
+	memset(&entries[0], 0, sizeof(entries));
+
+	va_start(va, ncolumns);
+
+	str += sprintf(str, "%s", margin);
+
+	for (i = 0; i < ncolumns; i++) {
+		struct table_entry *e = &entries[i];
+
+		e->width = va_arg(va, unsigned int);
+		e->next = l_strdup(va_arg(va, char*));
+
+		str += entry_append(e, str);
+	}
+
+	va_end(va);
+
+	display("%s\n", buf);
+	str = buf;
+
+	/*
+	 * The first column should now be indented, which effects the entry
+	 * width. Subtract this indentation only from the first column.
+	 */
+	entries[0].width -= strlen(margin) * 2;
+
+	while (!entries_done(ncolumns, &entries[0])) {
+		for (i = 0; i < ncolumns; i++) {
+			struct table_entry *e = &entries[i];
+
+			if (i == 0)
+				str += sprintf(str, "%s%s%s", margin,
+						margin, margin);
+
+			str += entry_append(e, str);
+		}
+
+		display("%s\n", buf);
+		str = buf;
+	}
+
+	for (i = 0; i < ncolumns; i++) {
+		if (entries[i].color)
+			l_free(entries[i].color);
+	}
+}
+
+void display_command_line(const char *command_family, const struct command *cmd)
 {
 	char *cmd_line = l_strdup_printf("%s%s%s%s%s%s%s",
 				command_family ? : "",
@@ -351,7 +563,7 @@ void display_command_line(const char *command_family,
 				cmd->arg ? " " : "",
 				cmd->arg ? : "");
 
-	display(MARGIN "%-*s%s\n", 50, cmd_line, cmd->desc ? : "");
+	display_table_row(MARGIN, 2, 50, cmd_line, 30, cmd->desc);
 
 	l_free(cmd_line);
 }
@@ -382,7 +594,7 @@ static void display_completion_matches(char **matches, int num_matches,
 		}
 
 		entry = l_strdup_printf("%-*s ", max_length, matches[index]);
-		strcpy(&line[line_used], entry);
+		l_strlcpy(&line[line_used], entry, sizeof(line) - line_used);
 		l_free(entry);
 
 		line_used += max_length + 1;
@@ -533,14 +745,21 @@ void display_enable_cmd_prompt(void)
 
 	rl_set_prompt(IWD_PROMPT);
 
-	display("");
+	/*
+	 * The following sequence of rl_* commands forces readline to properly
+	 * update its internal state and re-display the new prompt.
+	 */
+	rl_save_prompt();
+	rl_redisplay();
+	rl_restore_prompt();
+	rl_forced_update_display();
 }
 
 void display_disable_cmd_prompt(void)
 {
 	display_refresh_reset();
 
-	rl_set_prompt("Waiting to connect to IWD");
+	rl_set_prompt("Waiting for IWD to start...");
 	printf("\r");
 	rl_on_new_line();
 	rl_redisplay();
@@ -555,7 +774,7 @@ void display_agent_prompt(const char *label, bool mask_input)
 	if (mask_input)
 		reset_masked_input();
 
-	prompt = l_strdup_printf(COLOR_BLUE "%s " COLOR_OFF, label);
+	prompt = l_strdup_printf(COLOR_BLUE("%s "), label);
 
 	if (command_is_interactive_mode()) {
 		if (agent_saved_input) {
@@ -601,8 +820,8 @@ void display_agent_prompt_release(const char *label)
 
 	if (display_refresh.cmd) {
 		char *text = rl_copy_text(0, rl_end);
-		char *prompt = l_strdup_printf(COLOR_BLUE "%s " COLOR_OFF
-							"%s\n", label, text);
+		char *prompt = l_strdup_printf(COLOR_BLUE("%s ")
+						"%s\n", label, text);
 		l_free(text);
 
 		l_queue_push_tail(display_refresh.redo_entries, prompt);
@@ -626,42 +845,68 @@ void display_quit(void)
 	rl_crlf();
 }
 
-static void signal_handler(void *user_data)
+static void window_change_signal_handler(void *user_data)
 {
-	if (display_refresh.cmd)
-		display_refresh_reset();
+	display_refresh_check_feasibility();
 }
 
 static char *history_path;
 
 void display_init(void)
 {
-	const char *home_path;
+	const char *data_home;
+	char *data_path;
 
 	display_refresh.redo_entries = l_queue_new();
 
 	stifle_history(24);
 
-	home_path = getenv("HOME");
-	if (home_path)
-		history_path = l_strdup_printf("%s/%s", home_path,
-							".iwctl_history");
+	data_home = getenv("XDG_DATA_HOME");
+	if (!data_home || *data_home != '/') {
+		const char *home_path;
 
-	read_history(history_path);
+		home_path = getenv("HOME");
+		if (home_path)
+			data_path = l_strdup_printf("%s/%s/iwctl",
+						home_path, ".local/share");
+		else
+			data_path = NULL;
+	} else {
+		data_path = l_strdup_printf("%s/iwctl", data_home);
+	}
+
+	if (data_path) {
+		/*
+		 * If mkdir succeeds that means its a new directory, no need
+		 * to read the history since it doesn't exist
+		 */
+		if (mkdir(data_path, 0700) != 0) {
+			history_path = l_strdup_printf("%s/history", data_path);
+			read_history(history_path);
+		}
+
+		l_free(data_path);
+	} else {
+		history_path = NULL;
+	}
 
 	setlinebuf(stdout);
 
-	resize_signal = l_signal_create(SIGWINCH, signal_handler, NULL, NULL);
+	window_change_signal =
+		l_signal_create(SIGWINCH, window_change_signal_handler, NULL,
+									NULL);
 
 	rl_attempted_completion_function = command_completion;
 	rl_completion_display_matches_hook = display_completion_matches;
 
 	rl_completer_quote_characters = "\"";
 	rl_erase_empty_line = 1;
-	rl_callback_handler_install("Waiting for IWD to appear...",
+	rl_callback_handler_install("Waiting for IWD to start...",
 							readline_callback);
 
 	rl_redisplay();
+
+	display_refresh_check_feasibility();
 }
 
 void display_exit(void)
@@ -681,9 +926,11 @@ void display_exit(void)
 
 	l_io_destroy(io);
 
-	l_signal_remove(resize_signal);
+	l_signal_remove(window_change_signal);
+
+	if (history_path)
+		write_history(history_path);
 
-	write_history(history_path);
 	l_free(history_path);
 
 	display_quit();
